@@ -2,14 +2,54 @@
  * vim:ts=8:sw=3:sts=8:noexpandtab:cino=>5n-3f0^-2{2
  */
 
-#include "Eet_private.h"
-#include "Eet.h"
-
-#include <sys/types.h>
-#ifdef HAVE_SYS_MMAN_H
-# include <sys/mman.h>
+#ifdef HAVE_CONFIG_H
+# include <config.h>
 #endif
 
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
+#elif defined __GNUC__
+# define alloca __builtin_alloca
+#elif defined _AIX
+# define alloca __alloca
+#elif defined _MSC_VER
+# include <malloc.h>
+# define alloca _alloca
+#else
+# include <stddef.h>
+# ifdef  __cplusplus
+extern "C"
+# endif
+void *alloca (size_t);
+#endif
+
+#include <stdio.h>
+#include <errno.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/mman.h>
+#include <time.h>
+#include <string.h>
+#include <fnmatch.h>
+#include <fcntl.h>
+#include <unistd.h>
+#include <zlib.h>
+
+#ifdef HAVE_NETINET_IN_H
+# include <netinet/in.h>
+#endif
+
+#if defined(_WIN32) && ! defined(__CEGCC__)
+# include <winsock2.h>
+#endif
+
+#ifdef HAVE_EVIL
+# include <Evil.h>
+#endif
+
+#include "Eet.h"
+#include "Eet_private.h"
+
 #ifdef HAVE_REALPATH
 #undef HAVE_REALPATH
 #endif
@@ -17,6 +57,8 @@
 #define EET_MAGIC_FILE                  0x1ee7ff00
 #define EET_MAGIC_FILE_HEADER           0x1ee7ff01
 
+#define EET_MAGIC_FILE2                 0x1ee70f42
+
 typedef struct _Eet_File_Header         Eet_File_Header;
 typedef struct _Eet_File_Node           Eet_File_Node;
 typedef struct _Eet_File_Directory      Eet_File_Directory;
@@ -25,8 +67,10 @@ struct _Eet_File
 {
    char                 *path;
    FILE                 *fp;
+   FILE			*readfp;
    Eet_File_Header      *header;
    const unsigned char  *data;
+   Eet_Dictionary       *ed;
 
    int                   magic;
    int                   references;
@@ -53,19 +97,24 @@ struct _Eet_File_Directory
 
 struct _Eet_File_Node
 {
-   char          *name;
-   void          *data;
-   Eet_File_Node *next; /* FIXME: make buckets linked lists */
+   char                 *name;
+   void                 *data;
+   Eet_File_Node        *next; /* FIXME: make buckets linked lists */
 
-   int            offset;
-   int            size;
-   int            data_size;
+   int                   offset;
+   int                   dictionary_offset;
+   int                   name_offset;
 
-   unsigned char  free_name : 1;
-   unsigned char  compression : 1;
+   int                   name_size;
+   int                   size;
+   int                   data_size;
+
+   unsigned char         free_name : 1;
+   unsigned char         compression : 1;
 };
 
 #if 0
+/* Version 2 */
 /* NB: all int's are stored in network byte order on disk */
 /* file format: */
 int magic; /* magic number ie 0x1ee7ff00 */
@@ -78,17 +127,56 @@ struct
    int size; /* size of the data chunk */
    int data_size; /* size of the (uncompressed) data chunk */
    int name_size; /* length in bytes of the name field */
-   char name[name_size]; /* name string (variable length) */
+   char name[name_size]; /* name string (variable length) and \0 terminated */
 } directory[num_directory_entries];
 /* and now startes the data stream... */
 #endif
 
+#if 0
+/* Version 3 */
+/* NB: all int's are stored in network byte order on disk */
+/* file format: */
+int magic; /* magic number ie 0x1ee70f42 */
+int num_directory_entries; /* number of directory entries to follow */
+int num_dictionary_entries; /* number of dictionary entries to follow */
+struct
+{
+  int data_offset; /* bytes offset into file for data chunk */
+  int size; /* size of the data chunk */
+  int data_size; /* size of the (uncompressed) data chunk */
+  int name_offset; /* bytes offset into file for name string */
+  int name_size; /* length in bytes of the name field */
+  int flags; /* flags - for now 0 = uncompressed, 1 = compressed */
+} directory[num_directory_entries];
+struct
+{
+  int hash;
+  int offset;
+  int size;
+  int prev;
+  int next;
+} dictionary[num_dictionary_entries];
+/* now start the string stream. */
+/* and right after them the data stream. */
+#endif
+
+#define EET_FILE2_HEADER_COUNT                  3
+#define EET_FILE2_DIRECTORY_ENTRY_COUNT         6
+#define EET_FILE2_DICTIONARY_ENTRY_COUNT        5
+
+#define EET_FILE2_HEADER_SIZE                   (sizeof(int) * EET_FILE2_HEADER_COUNT)
+#define EET_FILE2_DIRECTORY_ENTRY_SIZE          (sizeof(int) * EET_FILE2_DIRECTORY_ENTRY_COUNT)
+#define EET_FILE2_DICTIONARY_ENTRY_SIZE         (sizeof(int) * EET_FILE2_DICTIONARY_ENTRY_COUNT)
+
 /* prototypes of internal calls */
 static Eet_File		*eet_cache_find(const char *path, Eet_File **cache, int cache_num);
 static void		eet_cache_add(Eet_File *ef, Eet_File ***cache, int *cache_num, int *cache_alloc);
 static void		eet_cache_del(Eet_File *ef, Eet_File ***cache, int *cache_num, int *cache_alloc);
 static int		eet_string_match(const char *s1, const char *s2);
+#if 0 /* Unused */
 static Eet_Error	eet_flush(Eet_File *ef);
+#endif
+static Eet_Error	eet_flush2(Eet_File *ef);
 static Eet_File_Node	*find_node_by_name(Eet_File *ef, const char *name);
 static int		read_data_from_disk(Eet_File *ef, Eet_File_Node *efn, void *buf, int len);
 
@@ -253,16 +341,194 @@ eet_cache_del(Eet_File *ef, Eet_File ***cache, int *cache_num, int *cache_alloc)
    *cache_alloc = new_cache_alloc;
 }
 
-/* internal string match. bails out at first mismatch - not comparing all */
-/* bytes in strings */
+/* internal string match. null friendly, catches same ptr */
 static int
 eet_string_match(const char *s1, const char *s2)
 {
    /* both null- no match */
    if ((!s1) || (!s2)) return 0;
+   if (s1 == s2) return 1;
    return (!strcmp(s1, s2));
 }
 
+/* flush out writes to a v2 eet file */
+static Eet_Error
+eet_flush2(Eet_File *ef)
+{
+   Eet_File_Node        *efn;
+   Eet_Error             error = EET_ERROR_NONE;
+   int                   head[EET_FILE2_HEADER_COUNT];
+   int                   num_directory_entries = 0;
+   int                   num_dictionary_entries = 0;
+   int                   bytes_directory_entries = 0;
+   int                   bytes_dictionary_entries = 0;
+   int                   bytes_strings = 0;
+   int                   data_offset = 0;
+   int                   strings_offset = 0;
+   int                   num;
+   int                   i;
+   int                   j;
+
+   if (eet_check_pointer(ef))
+     return EET_ERROR_BAD_OBJECT;
+   if (eet_check_header(ef))
+     return EET_ERROR_EMPTY;
+   if ((ef->mode != EET_FILE_MODE_WRITE) && (ef->mode != EET_FILE_MODE_READ_WRITE))
+     return EET_ERROR_NOT_WRITABLE;
+   if (!ef->writes_pending)
+     return EET_ERROR_NONE;
+
+   /* calculate string base offset and data base offset */
+   num = (1 << ef->header->directory->size);
+   for (i = 0; i < num; ++i)
+     {
+        for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
+          {
+             num_directory_entries++;
+             bytes_strings += strlen(efn->name) + 1;
+          }
+     }
+   if (ef->ed)
+     {
+        num_dictionary_entries = ef->ed->count;
+
+        for (i = 0; i < num_dictionary_entries; ++i)
+          bytes_strings += ef->ed->all[i].len;
+     }
+
+   /* calculate section bytes size */
+   bytes_directory_entries = EET_FILE2_DIRECTORY_ENTRY_SIZE * num_directory_entries + EET_FILE2_HEADER_SIZE;
+   bytes_dictionary_entries = EET_FILE2_DICTIONARY_ENTRY_SIZE * num_dictionary_entries;
+
+   /* calculate per entry offset */
+   strings_offset = bytes_directory_entries + bytes_dictionary_entries;
+   data_offset = bytes_directory_entries + bytes_dictionary_entries + bytes_strings;
+
+   for (i = 0; i < num; ++i)
+     {
+        for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
+          {
+             efn->offset = data_offset;
+             data_offset += efn->size;
+
+             efn->name_offset = strings_offset;
+             strings_offset += efn->name_size;
+          }
+     }
+
+   /* calculate dictionary strings offset */
+   if (ef->ed)
+     ef->ed->offset = strings_offset;
+
+   /* go thru and write the header */
+   head[0] = (int) htonl ((unsigned int) EET_MAGIC_FILE2);
+   head[1] = (int) htonl ((unsigned int) num_directory_entries);
+   head[2] = (int) htonl ((unsigned int) num_dictionary_entries);
+
+   fseek(ef->fp, 0, SEEK_SET);
+   if (fwrite(head, sizeof (head), 1, ef->fp) != 1)
+     goto write_error;
+
+   /* write directories entry */
+   j = 0;
+   for (i = 0; i < num; i++)
+     {
+        for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
+          {
+             int        ibuf[EET_FILE2_DIRECTORY_ENTRY_COUNT];
+
+             ibuf[0] = (int) htonl ((unsigned int) efn->offset);
+             ibuf[1] = (int) htonl ((unsigned int) efn->size);
+             ibuf[2] = (int) htonl ((unsigned int) efn->data_size);
+             ibuf[3] = (int) htonl ((unsigned int) efn->name_offset);
+             ibuf[4] = (int) htonl ((unsigned int) efn->name_size);
+             ibuf[5] = (int) htonl ((unsigned int) efn->compression);
+
+             if (fwrite(ibuf, sizeof(ibuf), 1, ef->fp) != 1)
+               goto write_error;
+          }
+     }
+
+   /* write dictionnary */
+   if (ef->ed)
+     {
+        int     offset = strings_offset;
+
+        for (j = 0; j < ef->ed->count; ++j)
+          {
+             int      sbuf[EET_FILE2_DICTIONARY_ENTRY_COUNT];
+
+             sbuf[0] = (int) htonl ((unsigned int) ef->ed->all[j].hash);
+             sbuf[1] = (int) htonl ((unsigned int) offset);
+             sbuf[2] = (int) htonl ((unsigned int) ef->ed->all[j].len);
+             sbuf[3] = (int) htonl ((unsigned int) ef->ed->all[j].prev);
+             sbuf[4] = (int) htonl ((unsigned int) ef->ed->all[j].next);
+
+             offset += ef->ed->all[j].len;
+
+             if (fwrite(sbuf, sizeof (sbuf), 1, ef->fp) != 1)
+               goto write_error;
+          }
+     }
+
+   /* write directories name */
+   for (i = 0; i < num; i++)
+     {
+        for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
+          {
+             if (fwrite(efn->name, efn->name_size, 1, ef->fp) != 1)
+               goto write_error;
+          }
+     }
+
+   /* write strings */
+   if (ef->ed)
+     {
+	for (j = 0; j < ef->ed->count; ++j)
+	  {
+	     if (ef->ed->all[j].str)
+	       {
+		  if (fwrite(ef->ed->all[j].str, ef->ed->all[j].len, 1, ef->fp) != 1)
+		    goto write_error;
+	       }
+	     else
+	       {
+		  if (fwrite(ef->ed->all[j].mmap, ef->ed->all[j].len, 1, ef->fp) != 1)
+		    goto write_error;
+	       }
+	  }
+     }
+
+   /* write data */
+   for (i = 0; i < num; i++)
+     {
+        for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
+          {
+             if (fwrite(efn->data, efn->size, 1, ef->fp) != 1)
+               goto write_error;
+          }
+     }
+
+   /* no more writes pending */
+   ef->writes_pending = 0;
+
+   return EET_ERROR_NONE;
+
+   write_error:
+   switch (ferror(ef->fp))
+     {
+      case EFBIG: error = EET_ERROR_WRITE_ERROR_FILE_TOO_BIG; break;
+      case EIO: error = EET_ERROR_WRITE_ERROR_IO_ERROR; break;
+      case ENOSPC: error = EET_ERROR_WRITE_ERROR_OUT_OF_SPACE; break;
+      case EPIPE: error = EET_ERROR_WRITE_ERROR_FILE_CLOSED; break;
+      default: error = EET_ERROR_WRITE_ERROR; break;
+     }
+   fclose(ef->fp);
+   ef->fp = NULL;
+   return error;
+}
+
+#if 0 /* Unused */
 /* flush out writes to an eet file */
 static Eet_Error
 eet_flush(Eet_File *ef)
@@ -354,7 +620,7 @@ eet_flush(Eet_File *ef)
 
    return EET_ERROR_NONE;
 
-write_error:
+   write_error:
    switch (ferror(ef->fp))
      {
       case EFBIG:
@@ -382,6 +648,7 @@ write_error:
    ef->fp = NULL;
    return EET_ERROR_WRITE_ERROR;
 }
+#endif
 
 EAPI int
 eet_init(void)
@@ -408,19 +675,17 @@ eet_clearcache(void)
    int	i;
 
    /*
-     We need to compute the list of eet file to close separately from the cache,
-     due to eet_close removing them from the cache after each call.
-   */
+    * We need to compute the list of eet file to close separately from the cache,
+    * due to eet_close removing them from the cache after each call.
+    */
    for (i = 0; i < eet_writers_num; i++)
      {
-	if (eet_writers[i]->references <= 0)
-	  num++;
+	if (eet_writers[i]->references <= 0) num++;
      }
 
    for (i = 0; i < eet_readers_num; i++)
      {
-	if (eet_readers[i]->references <= 0)
-	  num++;
+	if (eet_readers[i]->references <= 0) num++;
      }
 
    if (num > 0)
@@ -450,25 +715,205 @@ eet_clearcache(void)
 	  }
 
 	for (i = 0; i < num; i++)
-	  eet_close(closelist[i]);
+	  {
+	     eet_close(closelist[i]);
+	  }
      }
 }
 
-static Eet_File*
-eet_internal_read(Eet_File *ef)
+/* FIXME: MMAP race condition in READ_WRITE_MODE */
+static Eet_File *
+eet_internal_read2(Eet_File *ef)
+{
+   const int    *data = (const int*) ef->data;
+   const char   *start = (const char*) ef->data;
+   int           index = 0;
+   int           num_directory_entries;
+   int           bytes_directory_entries;
+   int           num_dictionary_entries;
+   int           bytes_dictionary_entries;
+   int           i;
+
+   index += sizeof(int);
+   if (eet_test_close((int) ntohl(*data) != EET_MAGIC_FILE2, ef))
+     return NULL;
+   data++;
+
+#define GET_INT(Value, Pointer, Index)          \
+   {                                            \
+      Value = ntohl(*Pointer);                  \
+      Pointer++;                                \
+      Index += sizeof(int);                     \
+   }
+
+   /* get entries count and byte count */
+   GET_INT(num_directory_entries, data, index);
+   /* get dictionary count and byte count */
+   GET_INT(num_dictionary_entries, data, index);
+
+   bytes_directory_entries = EET_FILE2_DIRECTORY_ENTRY_SIZE * num_directory_entries + EET_FILE2_HEADER_SIZE;
+   bytes_dictionary_entries = EET_FILE2_DICTIONARY_ENTRY_SIZE * num_dictionary_entries;
+
+   /* we cant have <= 0 values here - invalid */
+   if (eet_test_close((num_directory_entries <= 0), ef))
+     return NULL;
+
+   /* we cant have more bytes directory and bytes in dictionaries than the size of the file */
+   if (eet_test_close((bytes_directory_entries + bytes_dictionary_entries) > ef->data_size, ef))
+     return NULL;
+
+   /* allocate header */
+   ef->header = calloc(1, sizeof(Eet_File_Header));
+   if (eet_test_close(!ef->header, ef))
+     return NULL;
+
+   ef->header->magic = EET_MAGIC_FILE_HEADER;
+
+   /* allocate directory block in ram */
+   ef->header->directory = calloc(1, sizeof(Eet_File_Directory));
+   if (eet_test_close(!ef->header->directory, ef))
+     return NULL;
+
+   /* 8 bit hash table (256 buckets) */
+   ef->header->directory->size = 8;
+   /* allocate base hash table */
+   ef->header->directory->nodes = calloc(1, sizeof(Eet_File_Node *) * (1 << ef->header->directory->size));
+   if (eet_test_close(!ef->header->directory->nodes, ef))
+     return NULL;
+
+   /* actually read the directory block - all of it, into ram */
+   for (i = 0; i < num_directory_entries; ++i)
+     {
+        const char      *name;
+        Eet_File_Node   *efn;
+        int              name_offset;
+        int              name_size;
+        int              hash;
+
+        /* out directory block is inconsistent - we have oveerun our */
+        /* dynamic block buffer before we finished scanning dir entries */
+        efn = malloc (sizeof(Eet_File_Node));
+        if (eet_test_close(!efn, ef))
+          return NULL;
+
+        /* get entrie header */
+        GET_INT(efn->offset, data, index);
+        GET_INT(efn->size, data, index);
+        GET_INT(efn->data_size, data, index);
+        GET_INT(name_offset, data, index);
+        GET_INT(name_size, data, index);
+        GET_INT(efn->compression, data, index);
+
+#define EFN_TEST(Test, Ef, Efn)                 \
+        if (eet_test_close(Test, Ef))           \
+          {                                     \
+             free(Efn);                         \
+             return NULL;                       \
+          }
+
+        /* check data pointer position */
+        EFN_TEST(!((efn->size > 0)
+                   && (efn->offset + efn->size <= ef->data_size)
+                   && (efn->offset > bytes_dictionary_entries + bytes_directory_entries)), ef, efn);
+
+        /* check name position */
+        EFN_TEST(!((name_size > 0)
+                   && (name_offset + name_size < ef->data_size)
+                   && (name_offset >= bytes_dictionary_entries + bytes_directory_entries)), ef, efn);
+
+        name = start + name_offset;
+
+        /* check '\0' at the end of name string */
+        EFN_TEST(name[name_size - 1] != '\0', ef, efn);
+
+        efn->free_name = 0;
+        efn->name = (char*) name;
+        efn->name_size = name_size;
+
+        hash = _eet_hash_gen(efn->name, ef->header->directory->size);
+        efn->next = ef->header->directory->nodes[hash];
+        ef->header->directory->nodes[hash] = efn;
+
+        /* read-only mode, so currently we have no data loaded */
+        if (ef->mode == EET_FILE_MODE_READ)
+          efn->data = NULL;
+        /* read-write mode - read everything into ram */
+        else
+          {
+             efn->data = malloc(efn->size);
+             if (efn->data)
+               memcpy(efn->data, ef->data + efn->offset, efn->size);
+          }
+     }
+
+   ef->ed = NULL;
+
+   if (num_dictionary_entries)
+     {
+        const int       *dico = (const int*) ef->data + EET_FILE2_DIRECTORY_ENTRY_COUNT * num_directory_entries + EET_FILE2_HEADER_COUNT;
+        int              j;
+
+        if (eet_test_close((num_directory_entries * EET_FILE2_DICTIONARY_ENTRY_SIZE + index) > (bytes_dictionary_entries + bytes_directory_entries), ef))
+            return NULL;
+
+        ef->ed = calloc(1, sizeof (Eet_Dictionary));
+        if (eet_test_close(!ef->ed, ef)) return NULL;
+
+        ef->ed->all = calloc(num_dictionary_entries, sizeof (Eet_String));
+        if (eet_test_close(!ef->ed->all, ef)) return NULL;
+
+        ef->ed->count = num_dictionary_entries;
+	ef->ed->total = num_dictionary_entries;
+	ef->ed->start = start + bytes_dictionary_entries + bytes_directory_entries;
+	ef->ed->end = ef->ed->start;
+
+        for (j = 0; j < ef->ed->count; ++j)
+          {
+             int   hash;
+             int   offset;
+
+             GET_INT(hash, dico, index);
+             GET_INT(offset, dico, index);
+             GET_INT(ef->ed->all[j].len, dico, index);
+             GET_INT(ef->ed->all[j].prev, dico, index);
+             GET_INT(ef->ed->all[j].next, dico, index);
+
+             /* Hash value could be stored on 8bits data, but this will break alignment of all the others data.
+                So stick to int and check the value. */
+             if (eet_test_close(hash & 0xFFFFFF00, ef)) return NULL;
+
+             /* Check string position */
+             if (eet_test_close(!((ef->ed->all[j].len > 0)
+                                  && (offset > (bytes_dictionary_entries + bytes_directory_entries))
+                                  && (offset + ef->ed->all[j].len < ef->data_size)), ef))
+               return NULL;
+
+             ef->ed->all[j].mmap = start + offset;
+             ef->ed->all[j].str = NULL;
+
+	     if (ef->ed->all[j].mmap + ef->ed->all[j].len > ef->ed->end)
+	       ef->ed->end = ef->ed->all[j].mmap + ef->ed->all[j].len;
+
+             /* Check '\0' at the end of the string */
+             if (eet_test_close(ef->ed->all[j].mmap[ef->ed->all[j].len - 1] != '\0', ef)) return NULL;
+
+             if (ef->ed->all[j].prev == -1)
+               ef->ed->hash[hash] = j;
+          }
+     }
+
+   return ef;
+}
+
+static Eet_File *
+eet_internal_read1(Eet_File *ef)
 {
    const unsigned char	*dyn_buf = NULL;
    const unsigned char	*p = NULL;
-   int			index = 0;
-   int			num_entries;
-   int			byte_entries;
-   int			i;
-
-   if (eet_test_close((ef->data == (void *)-1) || (ef->data == NULL), ef))
-     return NULL;
-
-   if (eet_test_close(ef->data_size < sizeof(int) * 3, ef))
-     return NULL;
+   int			 index = 0;
+   int			 num_entries;
+   int			 byte_entries;
+   int			 i;
 
    /* build header table if read mode */
    /* geat header */
@@ -496,6 +941,10 @@ eet_internal_read(Eet_File *ef)
    if (eet_test_close((num_entries * 20) > byte_entries, ef))
      return NULL;
 
+   /* check we will not outrun the file limit */
+   if (eet_test_close(((byte_entries + sizeof(int) * 3) > ef->data_size), ef))
+     return NULL;
+
    /* allocate header */
    ef->header = calloc(1, sizeof(Eet_File_Header));
    if (eet_test_close(!ef->header, ef))
@@ -542,13 +991,15 @@ eet_internal_read(Eet_File *ef)
 	if (eet_test_close(!efn, ef))
 	  return NULL;
 
-	/* get entrie header */
+        /* get entrie header */
 	EXTRACT_INT(efn->offset, p, indexn);
 	EXTRACT_INT(efn->compression, p, indexn);
 	EXTRACT_INT(efn->size, p, indexn);
 	EXTRACT_INT(efn->data_size, p, indexn);
 	EXTRACT_INT(name_size, p, indexn);
 
+        efn->name_size = name_size;
+
 	/* invalid size */
 	if (eet_test_close(efn->size <= 0, ef))
 	  {
@@ -616,6 +1067,33 @@ eet_internal_read(Eet_File *ef)
    return ef;
 }
 
+static Eet_File *
+eet_internal_read(Eet_File *ef)
+{
+   const int    *data = (const int*) ef->data;
+
+   if (eet_test_close((ef->data == (void *)-1) || (ef->data == NULL), ef))
+     return NULL;
+
+   if (eet_test_close(ef->data_size < sizeof(int) * 3, ef))
+     return NULL;
+
+   switch (ntohl(*data))
+     {
+      case EET_MAGIC_FILE:
+	return eet_internal_read1(ef);
+      case EET_MAGIC_FILE2:
+	return eet_internal_read2(ef);
+      default:
+	ef->delete_me_now = 1;
+	eet_close(ef);
+	break;
+     }
+
+   return NULL;
+}
+
+#if 0 /* No prototype */
 EAPI Eet_File *
 eet_memopen_read(const void *data, size_t size)
 {
@@ -628,6 +1106,7 @@ eet_memopen_read(const void *data, size_t size)
    if (!ef)
      return NULL;
 
+   ef->ed = NULL;
    ef->path = NULL;
    ef->magic = EET_MAGIC_FILE;
    ef->references = 1;
@@ -641,6 +1120,7 @@ eet_memopen_read(const void *data, size_t size)
 
    return eet_internal_read(ef);
 }
+#endif
 
 EAPI Eet_File *
 eet_open(const char *file, Eet_File_Mode mode)
@@ -649,10 +1129,6 @@ eet_open(const char *file, Eet_File_Mode mode)
    Eet_File	*ef;
    struct stat	 file_stat;
 
-#ifdef _WIN32
-   HANDLE        h;
-#endif
-
    if (!file)
      return NULL;
 
@@ -663,7 +1139,8 @@ eet_open(const char *file, Eet_File_Mode mode)
 	ef = eet_cache_find((char *)file, eet_writers, eet_writers_num);
 	if (ef)
 	  {
-	     eet_flush(ef);
+	     eet_flush2(ef);
+	     ef->references++;
 	     ef->delete_me_now = 1;
 	     eet_close(ef);
 	  }
@@ -676,6 +1153,7 @@ eet_open(const char *file, Eet_File_Mode mode)
 	if (ef)
 	  {
 	     ef->delete_me_now = 1;
+	     ef->references++;
 	     eet_close(ef);
 	  }
 	ef = eet_cache_find((char *)file, eet_writers, eet_writers_num);
@@ -711,6 +1189,7 @@ eet_open(const char *file, Eet_File_Mode mode)
    if (ef && (file_stat.st_mtime != ef->mtime))
      {
 	ef->delete_me_now = 1;
+	ef->references++;
 	eet_close(ef);
 	ef = NULL;
      }
@@ -730,6 +1209,7 @@ eet_open(const char *file, Eet_File_Mode mode)
 
    /* fill some of the members */
    ef->fp = fp;
+   ef->readfp = NULL;
    ef->path = ((char *)ef) + sizeof(Eet_File);
    strcpy(ef->path, file);
    ef->magic = EET_MAGIC_FILE;
@@ -741,46 +1221,20 @@ eet_open(const char *file, Eet_File_Mode mode)
    ef->data = NULL;
    ef->data_size = 0;
 
+   /* FIXME: Add new ed on EET_FILE_MODE_WRITE */
+   ef->ed = mode == EET_FILE_MODE_WRITE ? eet_dictionary_add() : NULL;
+
    /* if we can't open - bail out */
    if (eet_test_close(!ef->fp, ef))
      return NULL;
 
-#ifndef _WIN32
    fcntl(fileno(ef->fp), F_SETFD, FD_CLOEXEC);
-#else
-   /* FIXME: check if that code is needed / correct */
-   h = (HANDLE) _get_osfhandle(fileno(ef->fp));
-   if (h == INVALID_HANDLE_VALUE)
-     return NULL;
-   if (!SetHandleInformation(h, HANDLE_FLAG_INHERIT, 0))
-     return NULL;
-#endif
    /* if we opened for read or read-write */
    if ((mode == EET_FILE_MODE_READ) || (mode == EET_FILE_MODE_READ_WRITE))
      {
-#ifdef _WIN32
-	HANDLE                  fm;
-#endif
-
-
 	ef->data_size = file_stat.st_size;
-#ifndef _WIN32
 	ef->data = mmap(NULL, ef->data_size, PROT_READ,
 			MAP_SHARED, fileno(ef->fp), 0);
-#else
-	fm = CreateFileMapping((HANDLE) _get_osfhandle (fileno(ef->fp)),
-			       NULL,
-			       PAGE_READONLY,
-			       0,
-			       0,
-			       NULL);
-	ef->data = MapViewOfFile(fm,
-				 FILE_MAP_READ,
-				 0,
-				 0,
-				 ef->data_size);
-	CloseHandle(fm);
-#endif
 
 	ef = eet_internal_read(ef);
 	if (!ef)
@@ -790,22 +1244,7 @@ eet_open(const char *file, Eet_File_Mode mode)
    /* we need to delete the original file in read-write mode and re-open for writing */
    if (ef->mode == EET_FILE_MODE_READ_WRITE)
      {
-	int i;
-
-	for (i = 0; i < ef->header->directory->size; i++)
-	  {
-	     Eet_File_Node      *efn;
-
-	     for (efn = ef->header->directory->nodes[i]; efn; efn = efn->next)
-	       {
-		  if (!efn->free_name)
-		    {
-		       efn->free_name = 1;
-		       efn->name = strdup(efn->name);
-		    }
-	       }
-	  }
-	fclose(ef->fp);
+	ef->readfp = ef->fp;
 	unlink(ef->path);
 	ef->fp = fopen(ef->path, "wb");
      }
@@ -846,7 +1285,7 @@ eet_close(Eet_File *ef)
    /* if its still referenced - dont go any further */
    if (ef->references > 0) return EET_ERROR_NONE;
    /* flush any writes */
-   err = eet_flush(ef);
+   err = eet_flush2(ef);
 
    /* if not urgent to delete it - dont free it - leave it in cache */
    if ((!ef->delete_me_now) && (ef->mode == EET_FILE_MODE_READ))
@@ -892,13 +1331,11 @@ eet_close(Eet_File *ef)
 	free(ef->header);
      }
 
-#ifndef _WIN32
-   if (ef->data) munmap((void*)ef->data, ef->data_size);
-#else
-   if (ef->data) UnmapViewOfFile (ef->data);
-#endif
+   eet_dictionary_free(ef->ed);
 
+   if (ef->data) munmap((void*)ef->data, ef->data_size);
    if (ef->fp) fclose(ef->fp);
+   if (ef->readfp) fclose(ef->readfp);
 
    /* zero out ram for struct - caution tactic against stale memory use */
    memset(ef, 0, sizeof(Eet_File));
@@ -1156,7 +1593,9 @@ eet_write(Eet_File *ef, const char *name, const void *data, int size, int compre
 	     return 0;
 	  }
 	efn->name = strdup(name);
-	efn->free_name = 1;
+        efn->name_size = strlen(efn->name) + 1;
+        efn->free_name = 1;
+
 	efn->next = ef->header->directory->nodes[hash];
 	ef->header->directory->nodes[hash] = efn;
 	efn->offset = 0;
@@ -1210,6 +1649,8 @@ eet_delete(Eet_File *ef, const char *name)
 	       ef->header->directory->nodes[hash] = efn->next;
 	     else
 	       pefn->next = efn->next;
+
+             if (efn->free_name) free(efn->name);
 	     free(efn);
 	     exists_already = 1;
 	     break;
@@ -1223,6 +1664,15 @@ eet_delete(Eet_File *ef, const char *name)
    return exists_already;
 }
 
+EAPI Eet_Dictionary *
+eet_dictionary_get(Eet_File *ef)
+{
+   if (eet_check_pointer(ef)) return NULL;
+
+   return ef->ed;
+}
+
+
 EAPI char **
 eet_list(Eet_File *ef, const char *glob, int *count_ret)
 {
